from datetime import UTC, date, datetime
from functools import reduce
from typing import Any, Callable, Generic, Literal, Protocol, Self, Tuple, Type, TypeVar
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy import DateTime, Select, delete, func, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

IDENTIFIABLE_SCHEMA_T = TypeVar("IDENTIFIABLE_SCHEMA_T")


class Identifiable(BaseModel, Generic[IDENTIFIABLE_SCHEMA_T]):
    id: UUID
    data: IDENTIFIABLE_SCHEMA_T


T_BASEMODEL = TypeVar("T_BASEMODEL", bound=BaseModel)


def recursive_get_dict(obj: Any) -> Any:
    if hasattr(obj, "__dict__"):
        return {
            k: recursive_get_dict(v) for k, v in obj.__dict__.items() if k[0] != "_"
        }
    elif isinstance(obj, list):
        return [recursive_get_dict(v) for v in obj]
    else:
        return obj


class BaseEntity(DeclarativeBase):

    @classmethod
    def from_basemodel(cls, mutation: T_BASEMODEL) -> "Self":
        return cls(**mutation.model_dump())

    def to_basemodel(self, model: Type[T_BASEMODEL]) -> T_BASEMODEL:
        return model.model_validate(recursive_get_dict(self))


def nowutc() -> datetime:
    return datetime.now(UTC)


class DatedEntity(BaseEntity):
    __abstract__ = True

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, default=nowutc
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, default=nowutc
    )


class IdentifiableEntity(BaseEntity):
    __abstract__ = True

    id: Mapped[UUID] = mapped_column(primary_key=True)

    @classmethod
    def from_identifiable(cls, model: Identifiable[T_BASEMODEL]) -> "Self":
        return cls(**{"id": model.id, **model.data.model_dump()})

    def to_identifiable(self, MODEL: Type[T_BASEMODEL]) -> Identifiable[T_BASEMODEL]:

        return Identifiable[MODEL].model_validate(  # type: ignore[valid-type]
            {"id": self.id, "data": recursive_get_dict(self)}
        )


IDENTIFIABLE_T = TypeVar("IDENTIFIABLE_T", bound=IdentifiableEntity)


class CRUDOperations(Generic[IDENTIFIABLE_T]):

    def __init__(
        self,
        entity_type: Type[IDENTIFIABLE_T],
        session_provider: Callable[[], AsyncSession],
    ) -> None:
        self.entity_type = entity_type
        self.session_provider = session_provider

    @property
    def session(self) -> AsyncSession:
        return self.session_provider()

    async def create(self, entity: IDENTIFIABLE_T) -> None:
        self.session.add(entity)
        await self.session.flush()
        await self.session.refresh(entity)

    async def get(self, id: UUID) -> IDENTIFIABLE_T:
        return await self.session.get_one(self.entity_type, id)

    async def update(self, entity: IDENTIFIABLE_T) -> None:
        await self.session.merge(entity)

    async def delete(self, entity: IDENTIFIABLE_T) -> None:
        await self.session.delete(entity)

    async def delete_by_id(self, id: UUID) -> None:
        await self.session.execute(
            delete(self.entity_type).where(self.entity_type.id == id)
        )

    async def update_by_id(self, id: UUID, entity: IDENTIFIABLE_T) -> None:
        await self.session.execute(
            update(self.entity_type)
            .where(self.entity_type.id == id)
            .values(entity.__dict__)
        )


QUERY_ENTITY_T = TypeVar("QUERY_ENTITY_T", bound=BaseEntity)
QUERY_FILTER_T = TypeVar("QUERY_FILTER_T")


class QueryInjector(Protocol):

    def inject(self, query: Select[Tuple[Any]], filter: Any) -> Select[Tuple[Any]]: ...


TRANSFORM_T = TypeVar("TRANSFORM_T")
PAGINATED_T = TypeVar("PAGINATED_T", bound=Any)


class Paginated(BaseModel, Generic[PAGINATED_T]):
    items: list[PAGINATED_T]
    total: int
    unfiltered_total: int

    def transform(
        self,
        transform: Callable[[PAGINATED_T], TRANSFORM_T],
    ) -> "Paginated[TRANSFORM_T]":
        return Paginated[TRANSFORM_T](
            items=[transform(item) for item in self.items],
            total=self.total,
            unfiltered_total=self.unfiltered_total,
        )


class QueryOperations(Generic[QUERY_FILTER_T, QUERY_ENTITY_T]):

    def __init__(
        self,
        entity_type: Type[QUERY_ENTITY_T],
        session_provider: Callable[[], AsyncSession],
        filters_functions: list[QueryInjector],
    ) -> None:
        self.entity_type: type[QUERY_ENTITY_T] = entity_type
        self.session_provider = session_provider
        self.filters_functions = filters_functions

    @property
    def session(self) -> AsyncSession:
        return self.session_provider()

    async def query(self, filter: QUERY_FILTER_T) -> "Paginated[QUERY_ENTITY_T]":
        unfiltered_total = (
            await self.session.execute(
                select(func.count()).select_from(self.entity_type)
            )
        ).scalar_one()

        filtered_query = self.generate_filtered_query(filter, select(self.entity_type))

        filtered_total = (
            await self.session.execute(
                select(func.count()).select_from(filtered_query.subquery())
            )
        ).scalar_one()

        return Paginated(
            items=[
                e
                for e in (
                    await self.session.execute(
                        self.generate_filtered_query(filter, select(self.entity_type))
                    )
                ).scalars()
            ],
            total=filtered_total,
            unfiltered_total=unfiltered_total,
        )

    def generate_filtered_query(
        self, filter: QUERY_FILTER_T, select_query: Select[Tuple[QUERY_ENTITY_T]]
    ) -> Select[Tuple[QUERY_ENTITY_T]]:
        return reduce(
            lambda query, filter_function: filter_function.inject(query, filter),
            self.filters_functions,
            select_query,
        )


# region PaginatedFilter
class PaginatedFilter(BaseModel):
    offset: int = 0
    limit: int = 10


class PaginatedQueryInjector(QueryInjector):
    def inject(
        self,
        query: Select[Tuple[BaseEntity]],
        filter: PaginatedFilter,
    ) -> Select[Tuple[BaseEntity]]:
        return query.offset(filter.offset).limit(filter.limit)


# endregion

# DateOrderedFilter


class DateOrderedFilter(BaseModel):
    order_by: Literal["asc", "desc"] = "asc"


class DateOrderedQueryInjector(QueryInjector):

    def __init__(self, entity_type: Type[DatedEntity]) -> None:
        self.entity_type = entity_type

    def inject(
        self,
        query: Select[Tuple[DatedEntity]],
        filter: DateOrderedFilter,
    ) -> Select[Tuple[DatedEntity]]:
        return query.order_by(getattr(self.entity_type.created_at, filter.order_by)())


# region Criteria


# region Criteria


class StringCriteria(BaseModel):
    value: str
    is_exact: bool
    case_sensitive: bool


class DateCriteria(BaseModel):
    value: date
    op: Literal["eq", "gt", "lt", "gte", "lte"]


class DatetimeCriteria(BaseModel):
    value: datetime
    op: Literal["eq", "gt", "lt", "gte", "lte"]


class CriteriaBasedAttributeQueryInjector(QueryInjector):

    def __init__(self, entity_type: Type[BaseEntity]) -> None:
        self.entity_type = entity_type

    def inject(
        self, query: Select[Tuple[BaseEntity]], filter: Any
    ) -> Select[Tuple[BaseEntity]]:

        attrs = filter.__dict__

        for field_name, value in attrs.items():

            if isinstance(value, (DateCriteria, DatetimeCriteria)):
                value = getattr(filter, field_name)

                entity_field = getattr(self.entity_type, field_name)

                op_mapping = {
                    "eq": entity_field == value.value,
                    "gt": entity_field > value.value,
                    "lt": entity_field < value.value,
                    "gte": entity_field >= value.value,
                    "lte": entity_field <= value.value,
                }

                query = query.filter(op_mapping[value.op])
            elif isinstance(value, StringCriteria):
                value = getattr(filter, field_name)

                entity_field = getattr(self.entity_type, field_name)

                if value.is_exact:
                    query = query.filter(entity_field == value.value)
                else:
                    query = query.filter(entity_field.contains(value.value))

        return query


# endregion



import asyncio
import logging
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Any, AsyncContextManager, AsyncGenerator, Callable

import uvloop
from croniter import croniter

from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.microservice import Microservice, SchedulerAppContext
from jararaca.scheduler.decorators import ScheduledAction

logger = logging.getLogger(__name__)


@dataclass
class SchedulerConfig:
    interval: int


class SchedulerBackend: ...


def extract_scheduled_actions(
    app: Microservice, container: Container
) -> list[tuple[Callable[..., Any], "ScheduledAction"]]:
    scheduled_actions: list[tuple[Callable[..., Any], "ScheduledAction"]] = []
    for controllers in app.controllers:

        controller_instance: Any = container.get_by_type(controllers)

        controller_scheduled_actions = ScheduledAction.get_type_scheduled_actions(
            controller_instance
        )
        scheduled_actions.extend(controller_scheduled_actions)

    return scheduled_actions


# TODO: Implement Backend for Distributed Lock
# TODO: Improve error handling
# TODO: Implement logging
# TODO: Implement tests
# TODO: Implement graceful shutdown
# TODO: Implement ScheduletAction parameters configuration
class Scheduler:

    def __init__(
        self,
        app: Microservice,
        backend: SchedulerBackend,
        config: SchedulerConfig,
    ) -> None:
        self.app = app
        self.backend = backend
        self.config = config
        self.container = Container(self.app)
        self.uow_provider = UnitOfWorkContextProvider(app, self.container)

        self.tasks: set[asyncio.Task[Any]] = set()
        self.lock = asyncio.Lock()
        self.shutdown_event = asyncio.Event()

        self.last_checks: dict[Callable[..., Any], datetime] = {}

        self.lifceycle = AppLifecycle(app, self.container)

    async def process_task(
        self, func: Callable[..., Any], scheduled_action: ScheduledAction
    ) -> None:

        async with self.lock:
            task = asyncio.create_task(self.handle_task(func, scheduled_action))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)

    async def handle_task(
        self, func: Callable[..., Any], scheduled_action: ScheduledAction
    ) -> None:

        last_check = self.last_checks.setdefault(func, datetime.now(UTC))

        cron = croniter(scheduled_action.cron, last_check)
        next_run: datetime = cron.get_next(datetime)
        if next_run > datetime.now(UTC):
            logger.info(
                f"Skipping {func.__module__}.{func.__qualname__} until {next_run}"
            )
            return

        action_specs = ScheduledAction.get_scheduled_action(func)

        assert action_specs is not None

        ctx: AsyncContextManager[Any]
        if action_specs.timeout:
            ctx = asyncio.timeout(action_specs.timeout)
        else:
            ctx = none_context()

        try:
            async with self.uow_provider(
                SchedulerAppContext(
                    action=func,
                    scheduled_to=next_run,
                    cron_expression=scheduled_action.cron,
                    triggered_at=datetime.now(UTC),
                )
            ):
                try:
                    async with ctx:
                        await func()
                except BaseException as e:
                    if action_specs.exception_handler:
                        action_specs.exception_handler(e)

        except Exception as e:
            logging.exception(f"Error in scheduled action {scheduled_action}: {e}")

        self.last_checks[func] = datetime.now(UTC)

    def run(self) -> None:

        async def run_scheduled_actions() -> None:

            async with self.lifceycle():
                scheduled_actions = extract_scheduled_actions(self.app, self.container)

                while True:
                    for func, scheduled_action in scheduled_actions:
                        if self.shutdown_event.is_set():
                            break

                        await self.process_task(func, scheduled_action)

                    await asyncio.sleep(self.config.interval)

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            runner.run(run_scheduled_actions())


@asynccontextmanager
async def none_context() -> AsyncGenerator[None, None]:
    yield



from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator

from fastapi import Depends, FastAPI, Request
from fastapi.params import Depends as DependsCls
from starlette.types import ASGIApp

from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.microservice import HttpAppContext
from jararaca.presentation.decorators import RestController
from jararaca.presentation.http_microservice import HttpMicroservice
from jararaca.presentation.websocket.websocket_interceptor import WebSocketInterceptor


class HttpAppLifecycle:

    def __init__(
        self,
        http_app: HttpMicroservice,
        lifecycle: AppLifecycle,
        uow_provider: UnitOfWorkContextProvider,
    ) -> None:
        self.lifecycle = lifecycle
        self.uow_provider = uow_provider
        self.http_app = http_app

    @asynccontextmanager
    async def __call__(self, api: FastAPI) -> AsyncGenerator[None, None]:
        async with self.lifecycle():

            websocket_interceptors = [
                interceptor
                for interceptor in self.lifecycle.initialized_interceptors
                if isinstance(interceptor, WebSocketInterceptor)
            ]

            for interceptor in websocket_interceptors:
                router = interceptor.get_ws_router(
                    self.lifecycle.app, self.lifecycle.container, self.uow_provider
                )

                api.include_router(router)

            for controller_t in self.lifecycle.app.controllers:
                controller = RestController.get_controller(controller_t)

                if controller is None:
                    continue

                instance: Any = self.lifecycle.container.get_by_type(controller_t)

                dependencies: list[DependsCls] = []
                for middleware in controller.middlewares:
                    middleware_instance = self.lifecycle.container.get_by_type(
                        middleware
                    )
                    dependencies.append(Depends(middleware_instance.intercept))

                router = controller.get_router_factory()(instance, dependencies)

                api.include_router(router)

                for middleware in self.http_app.middlewares:
                    middleware_instance = self.lifecycle.container.get_by_type(
                        middleware
                    )
                    api.router.dependencies.append(
                        Depends(middleware_instance.intercept)
                    )

            yield


class HttpUowContextProviderDependency:

    def __init__(self, uow_provider: UnitOfWorkContextProvider) -> None:
        self.uow_provider = uow_provider

    async def __call__(self, request: Request) -> AsyncGenerator[None, None]:
        async with self.uow_provider(HttpAppContext(request=request)):
            yield


def create_http_server(
    http_app: HttpMicroservice,
) -> ASGIApp:

    app = http_app.app
    factory = http_app.factory
    container = Container(app)

    uow_provider = UnitOfWorkContextProvider(app, container)
    http_uow_context_provider_dependency = HttpUowContextProviderDependency(
        uow_provider
    )

    lifespan = HttpAppLifecycle(
        http_app,
        AppLifecycle(app, container),
        uow_provider,
    )

    fastapi_app = (
        factory(lifespan) if factory is not None else FastAPI(lifespan=lifespan)
    )

    fastapi_app.router.dependencies.append(
        Depends(http_uow_context_provider_dependency)
    )

    return fastapi_app



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Generator, Protocol

from pydantic import BaseModel


class MessagePublisher(Protocol):
    async def publish(self, message: BaseModel, topic: str) -> None:
        raise NotImplementedError()


message_publishers_ctx = ContextVar[dict[str, MessagePublisher]](
    "message_publishers_ctx", default={}
)


@contextmanager
def provide_message_publisher(
    connection_name: str, message_publisher: MessagePublisher
) -> Generator[None, Any, None]:

    current_map = message_publishers_ctx.get({})

    token = message_publishers_ctx.set(
        {**current_map, connection_name: message_publisher}
    )

    try:
        yield
    finally:
        try:
            message_publishers_ctx.reset(token)
        except ValueError:
            pass


def use_publisher(connecton_name: str = "default") -> MessagePublisher:
    publisher = message_publishers_ctx.get({}).get(connecton_name)
    if publisher is None:
        raise ValueError(f"MessagePublisher not found for connection {connecton_name}")

    return publisher
